#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "tga.h"


#define DEBUG_ON

tga::~tga()
{
	if (data)
	free(data);
#ifdef DEBUG_ON
	printf("\nUnloaded\n");
#endif
}


tga::tga()
{
}


tga::tga(char *filename)
{
	load(filename);
}


void tga::load(char *filename)
{
	tga_header hdr;
	FILE *fp;
	int size;
	unsigned char temp;
	
	if ( (fp=fopen(filename,"rb")) == NULL)
	{
		perror("TGA_LOADER");
		exit(1);
	}
	
	fread(&hdr,sizeof(tga_header),1,fp);
			
	if (hdr.cmap_type)
	{
		printf("TGA_LOADER: Can't load colour-mapped Images\n");
		exit(1);
	}
	
	pixfmt=hdr.bpp >> 3;
	width=hdr.width;
	height=hdr.height;
	size=width*height*pixfmt;
	data=(byte *)malloc(size);
	//Position the pointer at the raw image data !
	fseek(fp,hdr.id_length,SEEK_CUR);	

	if(hdr.image_type > 3)
	{
		switch(pixfmt)
		{
		case 3: //24 bpp
			decode24bit(fp);
			break;
		case 1: //8 bpp
			decode8bit(fp);
			break;
		}
	}
	else ///Uncompressed data
	{
		fread(data,1,size,fp);
		if (pixfmt == 3)
		{
			//Convert BGR to RGB
			for(int i=0; i<size; i+=3)
			{
				temp=data[i];
				data[i]=data[i+2];
				data[i+2]=temp;
			}
		}
	}
#ifdef DEBUG_ON
	printf("Loaded %s: %dx%d @ %dbpp\n", filename, width,height,hdr.bpp);
#endif
	fclose(fp);
}


void tga::decode8bit(FILE *fp)
{
	byte packet, colour;
	byte *ptr;
	int count, offset=0, size;
	size=height*width;
	ptr=data;

	while(offset < size)
	{
		fread(&packet,1,1,fp);
		count=(packet&0x7f)+1;
		offset+=count;
		//Runlength packet
		if (packet&0x80)
		{
			//Read BGR and convert to RGB
			fread(&colour,1,1,fp);
			while(count)
			{
				*ptr=colour;
				ptr++;
				--count;
			}
		}
		else
		{
			//Raw packet
			while(count)
			{
				fread(&colour,1,1,fp);
				*ptr=colour;
				ptr++;
				--count;
			}
		}
	}
#ifdef DEBUG_ON
	printf("Offset=%d , Size=%d\n",offset,size);
#endif
}


void tga::decode24bit(FILE *fp)
{
	byte packet, colour[3];
	byte *ptr;
	int count, offset=0, size;
	size=height*width;
	ptr=data;

	//Sometimes offset exceeds size !
	while(offset < size)
	{
		fread(&packet,1,1,fp);
		count=(packet&0x7f)+1;
		offset+=count;
		//Runlength packet
		if (packet&0x80)
		{
			//Read BGR and convert to RGB
			fread(colour,3,1,fp);
			while(count)
			{
				ptr[0]=colour[2];
				ptr[1]=colour[1];
				ptr[2]=colour[0];
				ptr+=3;
				--count;
			}
		}
		else
		{
			//Raw packet
			while(count)
			{
				fread(colour,3,1,fp);
				ptr[0]=colour[2];
				ptr[1]=colour[1];
				ptr[2]=colour[0];
				ptr+=3;
				--count;
			}
		}
	}
#ifdef DEBUG_ON
	printf("Offset=%d , Size=%d\n",offset,size);
#endif
}


//Will be defined for future use!
void tga::decode32bit(FILE *fp)
{
}


void tga::write_24bpp(int w, int h, unsigned char *data, char *filename)
{
	FILE *fp;
	char *header=
		"\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00";
	//Open o/p file
	unsigned char info[6];
	unsigned char temp;
	unsigned char *copy;
	int size;
	
	size=w*h*3;

	info[0]=w;             //lo
	info[1]=(w>>8);        //hi 
	info[2]=h;             //lo
	info[3]=(h>>8);        //hi
	info[4]=24;
	info[5]=0;

	fp=fopen(filename,"wb");
	fwrite(header,1,12,fp);
	fwrite(info,1,6,fp);		
	
	//Just write the data in RGB fmt
	copy=(unsigned char *)malloc(size);
	memcpy(copy,data,size);
	//Convert this buffer to BGR fmt
	for(int i=0; i<size; i+=3)
	{
		temp=copy[i];
		copy[i]=copy[i+2];
		copy[i+2]=temp;
	}
	
	fwrite(copy,1,size,fp);
	fclose(fp);
	free(copy);
}



void tga::write_8bpp(int w, int h, unsigned char *data, char *filename)
{
	FILE *fp;
	char *header=
		"\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00";
	//Open o/p file
	unsigned char info[6];
	int size;
	
	size=w*h;

	info[0]=w;             //lo
	info[1]=(w>>8);        //hi 
	info[2]=h;             //lo
	info[3]=(h>>8);        //hi
	info[4]=8;
	info[5]=0;

	fp=fopen(filename,"wb");
	fwrite(header,1,12,fp);
	fwrite(info,1,6,fp);		
	
	fwrite(data,1,size,fp);
	fclose(fp);
}

// Slightly expensive but who cares..
// this is used in the preprocess stage
colour24 tga::getCol(int x, int y)
{
  colour24 a;
  int offset;
  while( y >= height)
    y=y-height;
  
  while( x >= width)
    x=x - width;
  
  offset=y*width+x;
  offset*=3;
  a.r=data[offset+1];
  a.g=data[offset+2];
  a.b=data[offset+3];
  return a;
}

